unit CommingSQLScript;

interface

uses
	Windows, SysUtils, Classes, DB, dbTables, BseMain, BtrMain, StrUtils ;

const
	DefaultTermChar  = '#';

type
	TcoScriptAction = (saFail, saAbort, saRetry, saIgnore, saContinue);

	TcoScriptErrorEvent = procedure(Sender: TObject; E: EDatabaseError;
		LineNo, StatementNo: Integer; var Action: TcoScriptAction) of object;

	TcoSQLScript = class(TComponent)
  private
		{ Private declarations }
		FSQL: TStrings;
		FParams: TParams;
		FQuery: TtbQuery;
		FTransaction: Boolean;
		FSemicolonTerm: Boolean;
		FIgnoreParams: Boolean;
		FTerm: Char;
		FBeforeExec: TNotifyEvent;
		FAfterExec: TNotifyEvent;
		FOnScriptError: TcoScriptErrorEvent;
		function GetSessionName: string;
		procedure SetSessionName(const Value: string);
		function GetDBSession: TttSession;
		function GetText: string;
		procedure ReadParamData(Reader: TReader);
		procedure WriteParamData(Writer: TWriter);
		function GetDatabase: TttDatabase;
		function GetDatabaseName: string;
		procedure SetDatabaseName(const Value: string);
		procedure CreateParams(List: TParams; const Value: PChar);
		procedure QueryChanged(Sender: TObject);
		procedure SetQuery(Value: TStrings);
		procedure SetParamsList(Value: TParams);
		function GetParamsCount: Cardinal;
	protected
		{ Protected declarations }
		procedure DefineProperties(Filer: TFiler); override;
		procedure CheckExecQuery(LineNo, StatementNo: Integer);
		procedure ExecuteScript(StatementNo: Integer); virtual;
	public
		{ Public declarations }
    constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;
    procedure ExecSQL;
    procedure ExecStatement(StatementNo: Integer);
		function ParamByName(const Value: string): TParam;
		property DBSession: TttSession read GetDBSession;
		property Text: string read GetText;
		property Database: TttDatabase read GetDatabase;
		property ParamCount: Cardinal read GetParamsCount;
	published
		{ Published declarations }
		property DatabaseName: string read GetDatabaseName write SetDatabaseName;
		property IgnoreParams: Boolean read FIgnoreParams write FIgnoreParams default False;
		property SemicolonTerm: Boolean read FSemicolonTerm write FSemicolonTerm default True;
		property SessionName: string read GetSessionName write SetSessionName;
		property Term: Char read FTerm write FTerm default DefaultTermChar;
		property SQL: TStrings read FSQL write SetQuery;
		property Params: TParams read FParams write SetParamsList {$IFDEF RX_D4} stored False {$ENDIF};
		property Transaction: Boolean read FTransaction write FTransaction;
		property BeforeExec: TNotifyEvent read FBeforeExec write FBeforeExec;
		property AfterExec: TNotifyEvent read FAfterExec write FAfterExec;
		property OnScriptError: TcoScriptErrorEvent read FOnScriptError write FOnScriptError;
	end;

procedure Register;

implementation

function IsLiteral(C: Char): Boolean;
begin
	Result := C in ['''', '"'];
end;

function NameDelimiter(C: Char; Delims: TCharSet): Boolean;
begin
	Result := (C in [' ', ',', ';', ')', #13, #10]) or (C in Delims);
end;

procedure CreateQueryParams(List: TParams; const Value: PChar; Macro: Boolean;
	SpecialChar: Char; Delims: TCharSet);
var
	CurPos, StartPos: PChar;
	CurChar: Char;
	Literal: Boolean;
	EmbeddedLiteral: Boolean;
	Name: string;

	function StripLiterals(Buffer: PChar): string;
	var
		Len: Word;
		TempBuf: PChar;

		procedure StripChar(Value: Char);
		begin
			if TempBuf^ = Value then
				StrMove(TempBuf, TempBuf + 1, Len - 1);
			if TempBuf[StrLen(TempBuf) - 1] = Value then
				TempBuf[StrLen(TempBuf) - 1] := #0;
		end;

	begin
		Len := StrLen(Buffer) + 1;
		TempBuf := AllocMem(Len);
		Result := '';
		try
			StrCopy(TempBuf, Buffer);
			StripChar('''');
			StripChar('"');
			Result := StrPas(TempBuf);
		finally
			FreeMem(TempBuf, Len);
		end;
	end;

begin
	if SpecialChar = #0 then Exit;
	CurPos := Value;
	Literal := False;
	EmbeddedLiteral := False;
	repeat
		CurChar := CurPos^;
		if (CurChar = SpecialChar) and not Literal and ((CurPos + 1)^ <> SpecialChar) then
		begin
			StartPos := CurPos;
			while (CurChar <> #0) and (Literal or not NameDelimiter(CurChar, Delims)) do begin
				Inc(CurPos);
				CurChar := CurPos^;
				if IsLiteral(CurChar) then begin
					Literal := Literal xor True;
					if CurPos = StartPos + 1 then EmbeddedLiteral := True;
				end;
			end;
			CurPos^ := #0;
			if EmbeddedLiteral then begin
				Name := StripLiterals(StartPos + 1);
				EmbeddedLiteral := False;
			end
			else Name := StrPas(StartPos + 1);
			if Assigned(List) then begin
				if List.FindParam(Name) = nil then begin
//					if Macro then
//						List.CreateParam(ftString, Name, ptInput).AsString := TrueExpr
//					else
					List.CreateParam(ftUnknown, Name, ptUnknown);
				end;
			end;
			CurPos^ := CurChar;
			StartPos^ := '?';
			Inc(StartPos);
			StrMove(StartPos, CurPos, StrLen(CurPos) + 1);
			CurPos := StartPos;
		end
		else if (CurChar = SpecialChar) and not Literal and ((CurPos + 1)^ = SpecialChar) then
			StrMove(CurPos, CurPos + 1, StrLen(CurPos) + 1)
		else if IsLiteral(CurChar) then Literal := Literal xor True;
		Inc(CurPos);
	until CurChar = #0;
end;

constructor TcoSQLScript.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
	FSQL := TStringList.Create;
	TStringList(SQL).OnChange := QueryChanged;
	FParams := TParams.Create(Self);
	FQuery := TtbQuery.Create(Self);
	FSemicolonTerm := True;
	FTerm := DefaultTermChar;
end;

destructor TcoSQLScript.Destroy;
begin
  FQuery.Free;
  FSQL.Free;
	FParams.Free;
  inherited Destroy;
end;

function TcoSQLScript.GetDatabase: TttDatabase;
begin
	Result := FQuery.Database;
end;

function TcoSQLScript.GetDatabaseName: string;
begin
  Result := FQuery.DatabaseName;
end;

procedure TcoSQLScript.SetDatabaseName(const Value: string);
begin
	FQuery.DatabaseName := Value;
end;

function TcoSQLScript.GetSessionName: string;
begin
	Result := FQuery.SessionName;
end;

procedure TcoSQLScript.SetSessionName(const Value: string);
begin
	FQuery.SessionName := Value;
end;

function TcoSQLScript.GetDBSession: TttSession;
begin
	Result := FQuery.DBSession;
end;

procedure TcoSQLScript.CheckExecQuery(LineNo, StatementNo: Integer);
var
	Done: Boolean;
	Action: TcoScriptAction;
	I: Integer;
	Param: TParam;
	Msg: array[0..255] of Char;
	S: string;
begin
	Done := False;
	repeat
		try
//			if IgnoreParams then FQuery.ExecDirect
//      else begin
				for I := 0 to FQuery.Params.Count - 1 do begin
					Param := FQuery.Params[I];
					Param.Assign(Params.ParamByName(Param.Name));
				end;
				FQuery.ExecSQL;
//			end;
			Done := True;
		except
			on E: EDatabaseError do begin
				Action := saFail;
//				S := Format(ResStr(SParseError), [ResStr(SMsgdlgError), LineNo]);
				if E is EDBEngineError then
					TDBError.Create(EDBEngineError(E), 0, LineNo, PChar(S) )
				else begin
					if E.Message <> '' then E.Message := E.Message + '. ';
					E.Message := E.Message + S;
				end;
				if Assigned(FOnScriptError) then
					FOnScriptError(Self, E, LineNo, StatementNo, Action);
				if Action = saFail then raise;
				if Action = saAbort then SysUtils.Abort;
				if Action = saContinue then begin
//					Application.HandleException(Self);
					Done := True;
				end
				else if Action = saIgnore then Done := True;
			end;
		end;
	until Done;
end;

procedure TcoSQLScript.ExecuteScript(StatementNo: Integer);
var
	S, LastStr: string;
	IsTrans, SQLFilled, StmtFound: Boolean;
	I, P, CurrStatement: Integer;
begin
//	IsTrans := FTransaction {$IFNDEF WIN32} and Database.IsSQLBased {$ENDIF}
//		and not TransActive(Database) and (StatementNo < 0);
	IsTrans := FTransaction and Database.IsSQLBased and (StatementNo < 0);
	LastStr := '';
	try
		if IsTrans then begin
//			if not Database.IsSQLBased then
//				Database.TransIsolation := tiDirtyRead;
			Database.StartTransaction;
		end;
	except
		IsTrans := False;
	end;
	try
		I := 0;
		CurrStatement := 0;
		StmtFound := False;
		while I < SQL.Count do begin
			FQuery.SQL.BeginUpdate;
			try
				FQuery.SQL.Clear;
				SQLFilled := False;
				repeat
					if LastStr <> '' then begin
						FQuery.SQL.Add(LastStr);
						LastStr := '';
					end;
					if I < SQL.Count then begin
						S := Trim(SQL[I]);
            Inc(I);
            P := Pos(';', S);
						if (P > 0) and FSemicolonTerm then begin
              LastStr := Trim(Copy(S, P + 1, MaxInt));
              S := Copy(S, 1, P - 1);
              if S <> '' then FQuery.SQL.Add(S);
							SQLFilled := True;
            end
            else begin
              if (S = Term) then SQLFilled := True
							else if S <> '' then FQuery.SQL.Add(S);
            end;
          end
					else SQLFilled := True;
        until SQLFilled;
      finally
        FQuery.SQL.EndUpdate;
			end;
      if FQuery.SQL.Count > 0 then begin
        if (StatementNo < 0) or (StatementNo = CurrStatement) then begin
					StmtFound := True;
          CheckExecQuery(I - 1, CurrStatement);
          if StatementNo = CurrStatement then Break;
        end;
        Inc(CurrStatement);
      end;
    end;
//		if not StmtFound then begin
//			DatabaseError(Format('%s: %d', [LoadStr(SListIndexError), StatementNo]));
//		end;
		if IsTrans then Database.Commit;
	except
		if IsTrans then Database.Rollback;
		raise;
	end;
end;

procedure TcoSQLScript.ExecStatement(StatementNo: Integer);
begin
//	if FSQL.Count = 0 then _DBError(SEmptySQLStatement);
//  FQuery.SetDBFlag(dbfExecScript, True);
	try
//    if not Database.Connected then _DBError(SDatabaseClosed);
		if Assigned(FBeforeExec) then FBeforeExec(Self);
		ExecuteScript(StatementNo);
		if Assigned(FAfterExec) then FAfterExec(Self);
	finally
//    FQuery.SetDBFlag(dbfExecScript, False);
	end;
end;

procedure TcoSQLScript.ExecSQL;
begin
  ExecStatement(-1);
end;

procedure TcoSQLScript.CreateParams(List: TParams; const Value: PChar);
begin
	CreateQueryParams(List, Value, False, ':', []);
end;

procedure TcoSQLScript.SetQuery(Value: TStrings);
begin
	TStringList(SQL).OnChange := nil;
  FSQL.Assign(Value);
	TStringList(SQL).OnChange := QueryChanged;
  QueryChanged(nil);
end;

function TcoSQLScript.GetText: string;
begin
	Result := SQL.Text;
end;

procedure TcoSQLScript.QueryChanged(Sender: TObject);
var
	List: TParams;
	P: PChar;
begin
	if not (csReading in ComponentState) then begin
		List := TParams.Create(Self);
		try
			CreateParams(List, PChar(Text));
			List.AssignValues(FParams);
			FParams.Clear;
			FParams.Assign(List);
		finally
				List.Free;
		end;
	end
	else begin
		FParams.Clear;
		CreateParams(FParams, PChar(Text));
	end;
end;

function TcoSQLScript.ParamByName(const Value: string): TParam;
begin
  Result := FParams.ParamByName(Value);
end;

procedure TcoSQLScript.SetParamsList(Value: TParams);
begin
	FParams.AssignValues(Value);
end;

function TcoSQLScript.GetParamsCount: Cardinal;
begin
	Result := FParams.Count;
end;

procedure TcoSQLScript.DefineProperties(Filer: TFiler);
begin
	inherited DefineProperties(Filer);
	Filer.DefineProperty('ParamData', ReadParamData, WriteParamData, True);
end;

procedure TcoSQLScript.ReadParamData(Reader: TReader);
begin
	Reader.ReadValue;
	Reader.ReadCollection(FParams);
end;

procedure TcoSQLScript.WriteParamData(Writer: TWriter);
begin
	Writer.WriteCollection(Params);
end;

procedure Register;
begin
	RegisterComponents('Comming PSQL', [TcoSQLScript]);
end;

end.




